/*=============================================================================
Vitalnix/ACCDB - Unified Account Database
backends/mig.c - Migration service back-end
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2004
  -- License restrictions apply (Dual FAL/LGPL2)
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  $Id: mig.c,v 1.5 2004/06/07 09:29:07 hirogen2 Exp $
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <libHX.h>
#include <accdb.h>
#include <accdb_int.h>

enum {
    CONF_READ = 0,
    CONF_FREE,
};

struct amod {
    struct accdb_module *info;
    int utrav, gtrav;
};

struct state {
    char *rd_str, *wd_str;
    struct HXlld *rd, *wd;
};

static void read_config(struct state *, int);

MODULE_NAME("Migration service back-end module");
MODULE_DESC("Transparent DB changer");
MODULE_INFO("Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2004");

//-----------------------------------------------------------------------------
int q_open(struct accdb_module *mp) {
    struct state *sp;
    char *br;
    int eax;

    if((sp = mp->state = calloc(1, sizeof(struct state))) == NULL) {
        return -errno;
    }

    read_config(sp, CONF_READ);

    if((sp->rd = HXlld_init(0)) == NULL || (sp->wd = HXlld_init(0)) == NULL) {
        goto __qopen_err;
    }

    // Open all specified back-end modules. If any of them fails, we fail.
    while((br = HX_strsep(&sp->rd_str, ",")) != NULL) {
        struct amod *md;
        if((md = malloc(sizeof(struct amod))) == NULL) { goto __qopen_err; }
        if((md->info = accdb_load(br)) == NULL) { free(md); goto __qopen_err; }
        md->utrav = md->gtrav = 0;
        HXlld_push(sp->rd, md);
    }

    // Specifying no WD back-end actually makes MIG a read-only DB
    while((br = HX_strsep(&sp->wd_str, ",")) != NULL) {
        struct amod *md;
        if((md = malloc(sizeof(struct amod))) == NULL) { goto __qopen_err; }
        if((md->info = accdb_load(br)) == NULL) { free(md); goto __qopen_err; }
        md->utrav = md->gtrav = 0;
        HXlld_push(sp->wd, md);
    }

    return 1;

__qopen_err:
    eax = -errno;
    q_close(mp);
    return eax;
}

//-----------------------------------------------------------------------------
int q_usertrav(struct accdb_module *mp, struct accdb_user *result) {
    struct state *sp = mp->state;
    struct HXlld_node *travp;

    if(result == NULL) {
        // Reset all traversers
        travp = sp->rd->first;
        while(travp != NULL) {
            struct amod *md = travp->ptr;
            md->info->usertrav(md->info->state, NULL);
            md->utrav = 0;
            travp = travp->Next;
        }

        travp = sp->wd->first;
        while(travp != NULL) {
            struct amod *md = travp->ptr;
            md->info->usertrav(md->info->state, NULL);
            md->utrav = 0;
            travp = travp->Next;
        }
        return 0;
    }

    /* We first search the WD back-end, it has the most recent users. Even
    though WD is a list, we only check the first, as all other WD back-ends are
    mirrors. */
    if(sp->wd->first != NULL) {
        struct amod *md = sp->wd->first->ptr;
        if(!md->utrav) {
            int rv = md->info->usertrav(md->info->state, result);
            if(rv != 0) { return rv; }
            md->utrav = 1;
        }
    }

    // Then come the RD back-ends
    travp = sp->rd->first;
    while(travp != NULL) {
        struct amod *md = travp->ptr;
        int rv;

        if(md->utrav) { travp = travp->Next; continue; }
        rv = md->info->usertrav(md->info->state, result);
        if(rv != 0) { return rv; }

        md->utrav = 1;
        travp = travp->Next;
    }

    return 0;
}

int q_userinfo(struct accdb_module *mp, struct accdb_user *rq,
 struct accdb_user *dest, size_t s) {
    struct state *sp = mp->state;
    struct HXlld_node *travp;
    size_t got = 0;

    if(sp->wd->first != NULL) {
        struct amod *md = sp->wd->first->ptr;
        md->info->userinfo(md->info->state, rq, dest, s);
    }

    return 0;
}

int q_useradd(struct accdb_module *mp, struct accdb_user *u) {
    return 0;
}

int q_usermod(struct accdb_module *mp, struct accdb_user *sm,
 struct accdb_user *mm) {
    return 0;
}

int q_userdel(struct accdb_module *mp, struct accdb_user *u) {
    return 0;
}

//-----------------------------------------------------------------------------
int q_grouptrav(struct accdb_module *mp, struct accdb_group *result) {
    struct state *sp = mp->state;
    struct HXlld_node *travp;

    if(result == NULL) {
        travp = sp->rd->first;
        while(travp != NULL) {
            struct amod *md = travp->ptr;
            md->info->grouptrav(md->info->state, NULL);
            md->gtrav = 0;
            travp = travp->Next;
        }

        travp = sp->wd->first;
        while(travp != NULL) {
            struct amod *md = travp->ptr;
            md->info->grouptrav(md->info->state, NULL);
            md->gtrav = 0;
            travp = travp->Next;
        }
        return 0;
    }

    if(sp->wd->first != NULL) {
        struct amod *md = sp->wd->first->ptr;
        if(md->gtrav) {
            int rv = md->info->grouptrav(md->info->state, result);
            if(rv != 0) { return rv; }
            md->gtrav = 1;
        }
    }

    travp = sp->rd->first;
    while(travp != NULL) {
        struct amod *md = travp->ptr;
        int rv;

        if(md->gtrav) { travp = travp->Next; continue; }
        rv = md->info->grouptrav(md->info->state, result);
        if(rv != 0) { return rv; }

        md->gtrav = 1;
        travp = travp->Next;
    }

    return 0;
}

int q_groupinfo(struct accdb_module *mp, struct accdb_group *rq,
 struct accdb_group *dest, size_t s) {
    return 0;
}

int q_groupadd(struct accdb_module *mp, struct accdb_group *g) {
    return 0;
}

int q_groupmod(struct accdb_module *mp, struct accdb_group *sm,
 struct accdb_group *mm) {
    return 0;
}

int q_groupdel(struct accdb_module *mp, struct accdb_group *g) {
    return 0;
}

//-----------------------------------------------------------------------------
int q_close(struct accdb_module *mp) {
    struct state *sp = mp->state;
    struct HXlld_node *travp;
    read_config(sp, CONF_FREE);

    if(sp->rd != NULL) {
        travp = sp->rd->first;
        while(travp != NULL) {
            struct amod *md = travp->ptr;
            md->info->close(md->info->state);
            accdb_unload(md->info);
            free(md);
            travp = travp->Next;
        }
        HXlld_free(sp->rd);
    }

    if(sp->wd != NULL) {
        travp = sp->wd->first;
        while(travp != NULL) {
            struct amod *md = travp->ptr;
            md->info->close(md->info->state);
            accdb_unload(md->info);
            free(md);
            travp = travp->Next;
        }
        HXlld_free(sp->wd);
    }

    return 0;
}

long q_sysctl(unsigned int req, ...) {
    int rv = 1;

    switch(req) {
        case ACCDB_ADDFLAGS:
        case ACCDB_DELFLAGS:
            rv = 0;
            break;
        default:
            rv = -ENOSYS;
            break;
    }

    return rv;
}

//-----------------------------------------------------------------------------
static void read_config(struct state *sp, int f) {
    const char *paths[] = {"backends/../vetc", "/etc", NULL};
    struct rconfig_opt table[] = {
        {"RD_LIST", RCONF_STRING, &sp->rd_str, NULL},
        {"WD_LIST", RCONF_STRING, &sp->wd_str, NULL},
        {NULL},
    };

    if(f == CONF_FREE) { HX_rconfig_free(table); }
    else { HX_rconfig_pv(paths, "libaccdb/mig", table, RCONFIG_ONE); }
    return;
}

//==[ End of file ]============================================================
