/*=============================================================================
Vitalnix/ACCDB - Unified Account Database
sysprog/groupadd.c - Create a new group
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/GPL2)
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: groupadd.c,v 1.12 2004/09/09 10:31:18 hirogen2 Exp $
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"

enum {
    E_SUCCESS = 0,
    E_OTHER,       // other error, see errno
    E_OPEN,        // unable to open back-end module or DB
    E_GID_USED,    // GID already used and -o was not specified
    E_NAME_USED,   // group already exists
    E_UPDATE,      // db->groupadd() did not return ok
    E_CLOSE,       // db->close() did not return ok
};

struct {
    long gid_min, gid_max;
} Conf = {
    gid_min: 100,
    gid_max: 60000,
};

struct {
    char *gname;
    long gid;

    // groupadd internal
    char *ac_after, *ac_before;
    int dup, inter, sys;
} Opt = {
    gname: NULL, gid: -1,
    ac_after: NULL, ac_before: NULL, dup: 0, inter: 0, sys: 0,
};

static int get_options(int *, const char ***);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct accdb_module *db;
    struct accdb_group mmask, smask;
    int eax, rv = E_SUCCESS;
    struct HX_repmap sr_map[] =
     {{'G', NULL, mmask.gname}, {'g', "%lu", &mmask.gid}, {0}};

    memset(&mmask, 0, sizeof(struct accdb_group));
    memset(&smask, 0, sizeof(struct accdb_group));
    read_config();
    if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

    // ----------------------------------------
    if((db = accdb_load(module_path)) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
         module_path, strerror(errno));
        return E_OPEN;
    }

    Opt.gname = *argv;
    if(Opt.inter) {
        struct itab t[] = {
            {"Group name\n[%s] > ", 's', &Opt.gname, NULL},
            {"Group ID (-1 for next free)\n[%d] > ", 'l', &Opt.gid, NULL},
            {NULL},
        };
        SH_interactive(t);
    }

    // ----------------------------------------
    if((eax = db->open(db)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        rv = E_OPEN;
        goto __main__close_accdb;
    }

    smask.gname = mmask.gname = Opt.gname;
    smask.gid = mmask.gid = -1;

    // ----------------------------------------
    if((eax = db->groupinfo(db, &smask, NULL, 0)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OTHER;
        goto __main__close_backend;
    } else if(eax > 0) {
        fprintf(stderr, "Group \"%s\" already exists\n", smask.gname);
        rv = E_NAME_USED;
        goto __main__close_backend;
    }

    // ----------------------------------------
    if(Opt.gid != -1) { // -g is provided
        smask.gname = NULL;
        smask.gid = Opt.gid;
        if(!Opt.dup && db->groupinfo(db, &smask, NULL, 0) > 0) {
            /* The -o flag (allow creating group with duplicate GID)
            was not passed. */
            fprintf(stderr, "Group with GID %ld already exists."
             " Use -o to override.\n", smask.gid);
            rv = E_GID_USED;
            goto __main__close_backend;
        }
        mmask.gid = Opt.gid;
    } else if(Opt.sys) { // -r flag passed
        mmask.gid = db->sysctl(ACCDB_NEXTGID_SYS, db);
    }

    // ----------------------------------------
    if(Opt.ac_before != NULL) { SH_runcmd(Opt.ac_before, sr_map); }

    if((eax = db->groupadd(db, &mmask)) <= 0) {
        fprintf(stderr, "Warning: Group addition failed: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        rv = E_UPDATE;
    } else if(Opt.ac_after != NULL) {
        SH_runcmd(Opt.ac_after, sr_map);
    }

    // ----------------------------------------
__main__close_backend:
    if((eax = db->close(db)) <= 0) {
        fprintf(stderr, "Warning: DB was not cleanly closed: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        // if there was some other error, keep that instead
        if(rv == E_SUCCESS) { rv = E_CLOSE; }
    }

__main__close_accdb:
    accdb_unload(db);
    return rv;
}

//-----------------------------------------------------------------------------
static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    const struct poptOption options_table[] = {
        // New, Vitalnix-groupadd options
        {NULL, 'A', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &Opt.ac_after,
         '\0', "Program to run after group addition", "cmd"},
        {NULL, 'B', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &Opt.ac_before,
         '\0', "Program to run before group addition", "cmd"},
        {NULL, 'I', POPT_ARG_NONE, &Opt.inter, '\0',
         "Interactively prompt for parameters", NULL},
        {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
         "Use a different module than \"*\" (the default)", "name"},

        // Default options
        {NULL, 'g', POPT_ARG_INT, &Opt.gid, '\0',
         "Numerical value of the group's ID", "gid"},
        {NULL, 'o', POPT_ARG_NONE, &Opt.dup, '\0',
         "Allow creating a group with non-unique GID", NULL},
        {NULL, 'r', POPT_ARG_NONE, &Opt.sys, '\0',
         "System group (use groupid < GID_MIN)", NULL},

        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    poptSetOtherOptionHelp(ctx, "[OPTIONS...] group");
    while((c = poptGetNextOpt(ctx)) >= 0);

    if(c < -1) {
        fprintf(stderr, "%s: %s: %s\n", **argv, poptBadOption(ctx, 0),
         poptStrerror(c));
        poptFreeContext(ctx);
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = SH_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
      *argc = 0;
      *argv = _empty_argv;
    }

    if(!Opt.inter && **argv == NULL) {
        // gname is mandatory
        poptPrintUsage(ctx, stderr, 0);
        poptFreeContext(ctx);
        return 0;
    }

    poptFreeContext(ctx);
    return 1;
}

static void read_config(void) {
    static const char *paths[] = {"sysprog/../vetc", "/etc", NULL};
    static struct rconfig_opt autouid_table[] = {
        {"GID_MIN", RCONF_LONG, &Conf.gid_min, NULL},
        {"GID_MAX", RCONF_LONG, &Conf.gid_max, NULL},
        {NULL},
    };
    static struct rconfig_opt options_table[] = {
        {"AC_BEFORE", RCONF_STRING, &Opt.ac_before, NULL},
        {"AC_AFTER",  RCONF_STRING, &Opt.ac_after,  NULL},
        {NULL},
    };

    HX_rconfig_pv(paths, "autouid", autouid_table, RCONFIG_ONE);
    HX_rconfig_pv(paths, "groupadd.conf", options_table, RCONFIG_ONE);
    return;
}

//==[ End of file ]============================================================
