/*=============================================================================
Vitalnix/ACCDB - Unified Account Database
sysprog/useradd.c - Create a new user
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/GPL2)
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: useradd.c,v 1.16 2004/09/09 10:31:18 hirogen2 Exp $
=============================================================================*/
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"

typedef unsigned char byte;

enum {
    E_SUCCESS = 0,
    E_OTHER,     // other error, see errno
    E_OPEN,      // unable to open back-end module or DB
    E_UID_USED,  // UID already used and -o was not specified
    E_NAME_USED, // user already exists
    E_UPDATE,    // db->useradd() did not return ok
    E_CREAT,     // post-stuff failed (home directory, skeleton copying)
    E_CLOSE,     // db->close() did not return ok
};

static struct {
    char *lname;
    long uid;
    char *igrp, *sgrp, *gecos, *homedir, *shell;

    char *passwd;
    long expire, inactive;

    // useradd internal
    char *ac_after, *ac_before, *skeldir;
    int creat, dup, force, inter, split_home, sys;
} Opt = {
    lname: NULL, uid: -1, igrp: NULL, sgrp: NULL, gecos: "(null)",
    homedir: NULL, shell: "/bin/bash", passwd: "!", expire: -1, inactive: -1,
    ac_after: NULL, ac_before: NULL, skeldir: "/etc/skel",
    creat: 0, dup: 0, force: 0, inter: 0, split_home: 0, sys: 0,
};

static struct {
    long uid_min, uid_max;
    // Defaults
    char *homebase;       // default homebase
    long keep_min, keep_max, warn_age;
    unsigned long umask;
} Conf = {
    uid_min: 100,
    uid_max: 60000,
    homebase: "/home",
    keep_min: 0, keep_max: 365, warn_age: 14,
    umask: 0066,
};

static void generate_homedir(void);
static int get_options(int *, const char ***);
static int interactive_lname(const char *, char, void *);
static void read_config(void);
static int pconfig_umask(const char *, byte, void *, void *);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    struct accdb_module *db;
    struct accdb_user mmask, smask;
    int eax, rv = E_SUCCESS;
    struct HX_repmap sr_map[] =
     {{'l', NULL, mmask.lname}, {'u', "%lu", &mmask.uid}, {0}};

    memset(&mmask, 0, sizeof(struct accdb_user));
    memset(&smask, 0, sizeof(struct accdb_user));
    read_config();
    if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }
    Opt.lname = *argv;

    // ----------------------------------------
    if((db = accdb_load(module_path)) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
         module_path, strerror(errno));
        return E_OPEN;
    }

    if(Opt.inter) {
        struct itab t[] = {
            {"Login name\n[%s] > ", 's', &Opt.lname, interactive_lname},
            {"User ID (-1 for next free)\n[%d] > ", 'l', &Opt.uid, NULL},
            {"Initial group (or GID)\n[%s] > ", 's', &Opt.igrp, NULL},
            {"Supplemental groups (and/or GIDs), seperated by comma\n[%s] > ",
             's', &Opt.sgrp, NULL},
            {"Comment field (real name, etc.)\n[%s] > ",
             's', &Opt.gecos, NULL},
            {"Home directory\n[%s] > ", 's', &Opt.homedir, NULL},
            {"Shell\n[%s] > ", 's', &Opt.shell, NULL},
            {"Skeleton dir\n[%s] > ", 's', &Opt.skeldir, NULL},
            {NULL},
        };
        SH_interactive(t);
    } else {
        if(Opt.homedir == NULL) { generate_homedir(); }
    }

    mmask.lname    = Opt.lname;
    mmask.uid      = Opt.uid;
    mmask.gecos    = Opt.gecos;
    mmask.home     = Opt.homedir;
    mmask.shell    = Opt.shell;

    mmask.passwd      = Opt.passwd;
    mmask.last_change = 0;
    mmask.keep_min    = Conf.keep_min;
    mmask.keep_max    = Conf.keep_max;
    mmask.warn_age    = Conf.warn_age;
    mmask.expire      = Opt.expire;
    mmask.inactive    = Opt.inactive;

    // ----------------------------------------
    if((eax = db->open(db)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        rv = E_OPEN;
        goto __main__close_accdb;
    }

    smask.lname = Opt.lname;
    smask.uid = smask.gid = -1;

    if((eax = db->userinfo(db, &smask, NULL, 0)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        rv = E_OTHER;
        goto __main__close_backend;
    } else if(eax > 0) {
        fprintf(stderr, "User \"%s\" already exists\n", smask.lname);
        rv = E_NAME_USED;
        goto __main__close_backend;
    }

    // ----------------------------------------
    if(Opt.uid != -1) { // -u is provided
        smask.lname = NULL;
        smask.uid = Opt.uid;
        smask.gid = -1;
        if(!Opt.dup && db->userinfo(db, &smask, NULL, 0) > 0) {
            /* The -o flag (allow creating user with duplicate UID)
            was not passed. */
            fprintf(stderr, "User with GID %ld already exists."
             " Use -o to override.\n", smask.gid);
            rv = E_UID_USED;
            goto __main__close_backend;
        }
        mmask.uid = Opt.uid;
    } else if(Opt.sys) { // -r flag passed
        mmask.uid = db->sysctl(ACCDB_NEXTUID_SYS, db);
    }

    if(SH_only_digits(Opt.igrp)) {
        mmask.gid  = strtoul(Opt.igrp, NULL, 0);
        mmask.igrp = NULL;
    } else {
        mmask.gid  = -1;
        mmask.igrp = Opt.igrp;
    }

    // ----------------------------------------
    if(Opt.ac_before != NULL) { SH_runcmd(Opt.ac_before, sr_map); }

    if((eax = db->useradd(db, &mmask)) <= 0) {
        fprintf(stderr, "Warning: User addition failed: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        rv = E_UPDATE;
        goto __main__close_backend;
    }

    if(Opt.creat) {
        if(HX_mkdir(Opt.homedir) <= 0) {
            fprintf(stderr, "Warning: Could not create home directory %s\n",
             strerror(errno));
            rv = E_CREAT;
            goto __main__close_backend;
        }

        lchown(Opt.homedir, mmask.uid, mmask.gid);
        chmod(Opt.homedir, 0755 & ~Conf.umask);

        if(Opt.skeldir != NULL) {
            HX_copy_dir(Opt.skeldir, Opt.homedir, HXF_UID | HXF_GID | HXF_KEEP,
             mmask.uid, mmask.gid);
        }
    }

    if(Opt.ac_after != NULL) {
        SH_runcmd(Opt.ac_after, sr_map);
    }

    // ----------------------------------------
__main__close_backend:
    if((eax = db->close(db)) <= 0) {
        fprintf(stderr, "Warning: DB was not cleanly closed: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        if(rv == E_SUCCESS) { rv = E_CLOSE; }
    }

__main__close_accdb:
    accdb_unload(db);
    return rv;
}

//-----------------------------------------------------------------------------
static void generate_homedir(void) {
    char buf[MAXFNLEN];
    if(Opt.lname == NULL || Conf.homebase == NULL) { return; }

    if(Opt.split_home == 1 || strlen(Opt.lname) == 1) {
        snprintf(buf, MAXFNLEN, "%s/%c/%s", Conf.homebase,
         *Opt.lname, Opt.lname);
    } else if(Opt.split_home == 2) {
        snprintf(buf, MAXFNLEN, "%s/%c/%c%c/%s", Conf.homebase,
         *Opt.lname, Opt.lname[0], Opt.lname[1], Opt.lname);
    } else {
        snprintf(buf, MAXFNLEN, "%s/%s", Conf.homebase, Opt.lname);
    }

    Opt.homedir = HX_strdup(buf);
    return;
}

static int get_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL, NULL};
    char *tmp;
    struct poptOption options_table[] = {
        // New, Vitalnix-useradd options
        {NULL, 'A', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &Opt.ac_after,
         '\0', "Program to run after user addition", "cmd"},
        {NULL, 'B', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &Opt.ac_before,
         '\0', "Program to run before user addition", "cmd"},
        {NULL, 'F', POPT_ARG_NONE, &Opt.force, '\0',
         "Force usage of dangerous umask", NULL},
        {NULL, 'I', POPT_ARG_NONE, &Opt.inter, '\0',
         "Interactively prompt for parameters", NULL},
        {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
         "Use a different module than \"*\" (the default)", "name"},
        {NULL, 'S', POPT_ARG_NONE, NULL, 'S',
         "Use split home feature (also -SS for two-level split)", NULL},

        // Default options
        {NULL, 'G', POPT_ARG_STRING, &Opt.sgrp, '\0',
         "Supplementary groups", "groups"},
        {NULL, 'c', POPT_ARG_STRING, &Opt.gecos, '\0',
         "User's comment in the password file", NULL},
        {NULL, 'd', POPT_ARG_STRING, &Opt.homedir, '\0',
         "Home directory of the user (overrides -S)", "dir"},
        {NULL, 'e', POPT_ARG_STRING, &tmp, 'e',
         "Date when the account expires", "date"},
        {NULL, 'f', POPT_ARG_INT, &Opt.inactive, '\0',
         "Days until account becomes inactive", "days"},
        {NULL, 'g', POPT_ARG_STRING, &Opt.igrp, '\0',
         "Initial group of the user", "group"},
        {NULL, 'k', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &Opt.skeldir,
         '\0', "Skeleton directory", "dir"},
        {NULL, 'm', POPT_ARG_NONE, &Opt.creat, '\0',
         "Create home directory", NULL},
        {NULL, 'o', POPT_ARG_NONE, &Opt.dup, '\0',
         "Allow creating a user with non-unique UID", NULL},
        {NULL, 'p', POPT_ARG_STRING, &Opt.passwd, '\0',
         "Encrypted password to use", NULL},
        {NULL, 'r', POPT_ARG_NONE, &Opt.sys, '\0',
         "System account (use userid < UID_MIN)", NULL},
        {NULL, 's', POPT_ARG_STRING, &Opt.shell, '\0',
         "The user's default shell", "prog"},
        {NULL, 'u', POPT_ARG_LONG, &Opt.uid, '\0',
         "Numerical value of the user's ID", "uid"},

        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0) {
        switch(c) {
            case 'S':
                if(Opt.split_home < 2) { ++Opt.split_home; }
                break;
            case 'e':
                Opt.expire = SH_string_to_iday(tmp);
                break;
        }
    }

    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        poptFreeContext(ctx);
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = SH_count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    if(!Opt.inter && (args == NULL || *args == NULL)) {
        // loginname is mandatory
        poptPrintUsage(ctx, stderr, 0);
        poptFreeContext(ctx);
        fprintf(stderr, "You need to specify a username\n");
        return 0;
    }

    poptFreeContext(ctx);
    return 1;
}

static int interactive_lname(const char *key, char type, void *ptr) {
    char **lname = ptr;
    if(**lname == '\0') { return 0; }
    if(Opt.homedir == NULL) { generate_homedir(); }
    return 1;
}

static void read_config(void) {
    static const char *paths[] = {"sysprog/../vetc", "/etc", NULL};
    static struct rconfig_opt autouid_table[] = {
        {"UID_MIN", RCONF_LONG, &Conf.uid_min, NULL},
        {"UID_MAX", RCONF_LONG, &Conf.uid_max, NULL},
        {NULL},
    };
    static struct rconfig_opt options_table[] = {
        {"AC_BEFORE", RCONF_STRING, &Opt.ac_before, NULL},
        {"AC_AFTER",  RCONF_STRING, &Opt.ac_after,  NULL},

        {"GROUP",       RCONF_STRING, &Opt.igrp,       NULL},
        {"HOME",        RCONF_STRING, &Conf.homebase,  NULL},
        {"CREATE_HOME", RCONF_IBOOL,  &Opt.creat,      NULL},
        {"SPLIT_HOME",  RCONF_IBOOL,  &Opt.split_home, NULL},
        {"UMASK",       RCONF_ULONG,  &Conf.umask,     pconfig_umask},
        {"SKEL",        RCONF_STRING, &Opt.skeldir,    NULL},
        {"SHELL",       RCONF_STRING, &Opt.shell,      NULL},

        {"PASS_KEEP_MIN", RCONF_ULONG, &Conf.keep_min, NULL},
        {"PASS_KEEP_MAX", RCONF_ULONG, &Conf.keep_max, NULL},
        {"PASS_WARN_AGE", RCONF_ULONG, &Conf.warn_age, NULL},
        {"EXPIRE",        RCONF_LONG,  &Opt.expire,    NULL},
        {"INACTIVE",      RCONF_LONG,  &Opt.inactive,  NULL},
        {NULL},
    };

    HX_rconfig_pv(paths, "autouid", autouid_table, RCONFIG_ONE);
    HX_rconfig_pv(paths, "useradd.conf", options_table, RCONFIG_ONE);
    return;
}

static int pconfig_umask(const char *key, byte type, void *ptr, void *uptr) {
    unsigned long *mask = ptr;
    if((*mask & (S_IWGRP | S_IWOTH)) != (S_IWGRP | S_IWOTH) && !Opt.force) {
        // complain if not all write permission bits are cleared
        fprintf(stderr, "Error: will refuse to allow foreign write-access for"
         " user directory,\n" "use -F to override.\n");
    } else if((*mask & (S_IRGRP | S_IROTH)) != (S_IRGRP | S_IROTH) &&
     !Opt.force) {
        // warn if not all read permission bits are cleared
        fprintf(stderr, "Warning: other users will possibly be able to spy on"
         " the home directory\n");
    }
    return 1;
}

//==[ End of file ]============================================================
